from pydoc import locate
from datetime import datetime
from pg_cnab240.payment import Payment
import os
import random


class File:
    def __init__(self, bank, company=None, payments=[]):
        self.company = company
        self.payments = payments

        self.bank = self.import_bank(bank)
        self.header = self.import_header()
        self.footer = self.import_footer()
        self.body = []
        self.lots_quantity = 1
        self.lines = []
        self.line_cursor = 0
    
    def import_bank(self, bank):
        bankClassFile =  locate('pg_cnab240.banks.' + bank + '.' + bank)
        bankClass = getattr(bankClassFile, bank)
        return bankClass()

    def import_header(self):
        return self.bank.get_file_header()
    
    def import_footer(self):
        return self.bank.get_file_footer()
    
    def verify(self):
        if self.header is None or self.footer is None:
            raise Exception('Header and Footer cannot be None')
        return True
    
    def add_payment(self, payment):
        self.payments.append(payment)
    
    def process_payments(self):
        for payment in self.payments:
            register_number = 1
            
            # get bank payment segment
            payment_segment = self.bank.get_payment_segment(payment.get_attribute('type'))
            segment = payment_segment['segment_class']()

            # set bank and company
            segment.set_bank(self.bank)
            segment.set_company(self.company)

            # attr payment attributes
            segment_data = payment.attributes
            segment_data['payment_way'] = payment_segment['payment_types'][payment.get_attribute('type')]
            
            # set custom segment attributes
            segment_data['lot_code'] = self.lots_quantity
            segment_data['register_number'] = register_number
            segment.set_data(segment_data)

            # check header
            if hasattr(segment, 'get_header_line'):
                self.body.append(segment.get_header_line())
            
            self.body.append(segment.to_line())

            # check footer
            if hasattr(segment, 'get_footer_line'):
                self.body.append(segment.get_footer_line())

            # increment lots_quantity
            self.lots_quantity += 1

    def generate(self, file_path=None, file_name=None):
        self.verify()

        # add header line
        self.lines.append(self.header.to_line())

        # process payments
        self.process_payments()

        for line in self.body:
            self.lines.append(line)

        # populate footer
        self.footer.set_data(dict(
            lots_quantity = self.lots_quantity - 1,
            registers_quantity = ((self.lots_quantity - 1) * 3) + 2,
        ))

        # add footer line
        self.lines.append(self.footer.to_line())

        if file_path:
            return self.save_file(file_path, file_name)

        return True
    
    def read_file_content(self, file_content=""):
        if not file_content:
            raise Exception('File content cannot be empty.')
        
        # split file in lines
        self.lines = file_content.splitlines()

        line_index = 0
        payment_header_line = None
        for line in self.lines:
            if line_index == 0:
                # header
                self.header.set_attributes_from_line(line)
                if hasattr(self.header, 'extract_company'):
                    self.company = self.header.extract_company()
            elif line_index == (len(self.lines) - 1):
                # footer
                self.footer.set_attributes_from_line(line)
                pass
            else:
                # payments
                if line[8] == 'C':
                    payment_header_line = line
                if line[8] != 'C' and line[8] != ' ':
                    # get payment segment data
                    payment_segment_data = self.bank.identify_payment_segment(line, payment_header_line)
                    if not payment_segment_data:
                        raise Exception('Cannot identify payment segment: ' + payment_header_line + ', ' + line)
                    
                    # get segment class
                    segment_class = payment_segment_data['segment_class']
                    payment_segment = segment_class()
                    payment_segment.set_attributes_from_line(line)
                    
                    payment = Payment()
                    for attr_name, attr_value in payment_segment.get_dict().items():
                        payment.set_attribute(attr_name, attr_value)
                    
                    self.payments.append(payment)
            
            line_index += 1
    
    def get_content(self):
        content = ""
        for line in self.lines:
            if content:
                content += "\n"
            content += line
        return content
    
    def save_file(self, file_path, file_name):
        # check file name
        if not file_name:
            file_name = datetime.utcnow().strftime("%Y-%m-%d_%H-%M-%S_") + str(random.randint(0, 10000) * 5) + '.rem'
        
        file_full_path = os.path.join(file_path, file_name)
        f = open(file_full_path, 'w')
        f.write(self.get_content())
        f.close()

        return file_full_path
    
    def next_line(self, append_line_break=False):
        if self.line_cursor > (len(self.lines) - 1):
            return None
        
        line = ""

        if append_line_break and self.line_cursor > 0:
            line += "\n"
        
        line += self.lines[self.line_cursor]
        
        self.line_cursor += 1
        return line
    
    def read(self, file_content):
        self.verify()
        pass



class BankAccount:
    def __init__(self, bank_code, agency, account, digit):
        self.bank_code = bank_code
        self.agency = agency
        self.account = account
        self.digit = digit



from pg_cnab240.banks.bank import Bank
from pg_cnab240.banks.itau.file_header import FileHeader
from pg_cnab240.banks.itau.file_footer import FileFooter
from pg_cnab240.banks.itau.segments.A import SegmentA, SegmentANF
from pg_cnab240.banks.itau.segments.J import SegmentJ


class itau(Bank):
    def __init__(self):
        super().__init__('Itaú', 'itau', 341, 13, 1, 'payment_way')
        super().set_segment('J', SegmentJ, {
            'slip': '30',
            'other_bank_slip': '31',
        })
        super().set_segment('A', SegmentA, {
            'cc': '01',
            'admin_check': '02',
            'doc': '03',
            'poupanca': '05',
            'ccd': '06',
            'docd': '07',
            'normal_darf': '16',
            'simple_darf': '18',
            'gps': '17',
            'city_taxs': '19',
            'darj': '21',
            'gare': '22',
            'ipva': '25',
            'dpvat': '27',
            'billing_title': '30',
            'billing_title_other_bank': '31',
            'fiscal_note': '32',
            'fgts': '35',
            'ted': '41',
            'tedd': '43',
            'gnre': '91',
        })
        super().set_segment('ANF', SegmentANF, {
            'nf': None,
        }, 'A')
    
    def get_file_header(self):
        file_header = FileHeader()
        file_header.set_bank(self)
        return file_header
    
    def get_file_footer(self):
        file_footer = FileFooter()
        file_footer.set_bank(self)
        return file_footer
    
    def get_company_document_id(self, document_type):
        if document_type == 'cnpj':
            return 2
        return 1 # cpf



import unittest
from pg_cnab240.file import File
from pg_cnab240.company import Company
from pg_cnab240.payment import Payment
from datetime import datetime
import os

# create company object
company = Company('BF Servicos De Cobranca Ltda', '07179434000140')
company.set_bank_acccount(341, '00772', '69637', 3)
company.set_address('Av. Andrômeda', 2000, 'Bl.8-4 andar', 'Alphaville Residencial Plus', 'Barueri', 'SP', '06473000')

class ItauFileTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass
    
    def test_1_generate_header(self):
        payment_file = File('itau', company)
        payment_file.header.set_company_data(company)
        header_line = payment_file.header.to_line()
        # print(header_line)
        assert '240' in str(len(header_line))
    
    def test_2_process_one_paymentA(self):
        payment_file = File('itau', company)
        payment_file.header.set_company_data(company)
        payment_file.payments = []

        payment = Payment(type='ted', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, your_number='5511972063805440', pay_date='10052018', ispb_code='90400888', payment_amount=2400.00, favored_document_number='11111111111111')

        # add payment
        payment_file.add_payment(payment)

        # generate
        payment_file.generate()

        body_big_line = ''.join(payment_file.body)

        root_path = os.path.dirname(os.path.abspath(__file__))
        body_big_file_path = os.path.join(root_path, 'body_big_file_segment_a.txt')
        f = open(body_big_file_path, 'w')
        f.write(body_big_line)
        f.close()

        assert '34100011C2041040 207179434000140                    01111 000011111111 3BF SERVICOS DE COBRANCA LTDA                                          AV ANDROMEDA                  02000BL84 ANDAR     BARUERI             06473000SP                  3410001300001A00000003301111 000011111111 0CLIENTE TESTE                 5511972063805440    10052018REA904008880000000000000000240000                            000000000000000                    00000011111111111111                       34100015         000003000000000000240000000000000000000000                                                                                                                                                                                     ' in body_big_line
    
    def test_3_process_one_paymentJ(self):
        payment_file = File('itau', company)
        payment_file.header.set_company_data(company)
        payment_file.payments = []

        payment = Payment(type='other_bank_slip', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, pay_date='10052018', ispb_code='90400888', currency_type=9, dv='4', due_rule='7524', amount=10067.60, free_field='9485239700000007661190101', due_date='14052018', title_amount=10067.60, payment_amount=10067.60, your_number='5506715023835136')

        # add payment
        payment_file.add_payment(payment)

        # generate
        payment_file.generate()

        body_big_line = None
        body_big_line = ''.join(payment_file.body)

        root_path = os.path.dirname(os.path.abspath(__file__))
        body_big_file_path = os.path.join(root_path, 'body_big_file_segment_j.txt')
        f = open(body_big_file_path, 'w')
        f.write(body_big_line)
        f.close()

        assert '34100011C2031030 207179434000140                    01111 000011111111 3BF SERVICOS DE COBRANCA LTDA                                          AV ANDROMEDA                  02000BL84 ANDAR     BARUERI             06473000SP                  3410001300001J00003394752400010067609485239700000007661190101CLIENTE TESTE                 14052018000000001006760000000000000000000000000000000100520180000000010067600000000000000005506715023835136                                          34100015         000003000000000001006760000000000000000000                                                                                                                                                                                     ' in body_big_line
    
    def test_4_process_one_paymentANF(self):
        payment_file = File('itau', company)
        payment_file.header.set_company_data(company)
        payment_file.payments = []

        payment = Payment(type='nf', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, pay_date='10052018', ispb_code='90400888', payment_amount=10067.60, dv='4', due_rule='7524', amount=10067.60, free_field='9485239700000007661190101', due_date='14052018', title_amount=10067.60, your_number='5511972063805440', nf_document='5646510065784320', favored_document_number='11111111111111')

        # add payment
        payment_file.add_payment(payment)

        # generate
        payment_file.generate()

        body_big_line = None
        body_big_line = ''.join(payment_file.body)

        root_path = os.path.dirname(os.path.abspath(__file__))
        body_big_file_path = os.path.join(root_path, 'body_big_file_segment_anf.txt')
        f = open(body_big_file_path, 'w')
        f.write(body_big_line)
        f.close()

        assert '34100011C2041040 207179434000140                    01111 000011111111 3BF SERVICOS DE COBRANCA LTDA                                          AV ANDROMEDA                  02000BL84 ANDAR     BARUERI             06473000SP                  3410001300001A00000003301111 000011111111 0CLIENTE TESTE                 5511972063805440    10052018REA000000000000000000000001006760                            00000000000000056465100657843      000000111111111111112                      34100015         000003000000000001006760000000000000000000                                                                                                                                                                                     ' in body_big_line
    
    def test_5_process_many_payments(self):
        payment_file = File('itau', company)
        payment_file.header.set_company_data(company)
        payment_file.payments = []

        paymentA = Payment(type='ted', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, your_number='5511972063805440', pay_date='10052018', ispb_code='90400888', payment_amount=2400.00, favored_document_number='11111111111111')

        paymentJ = Payment(type='other_bank_slip', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, pay_date='10052018', ispb_code='90400888', currency_type=9, dv='4', due_rule='7524', amount=10067.60, free_field='9485239700000007661190101', due_date='14052018', title_amount=10067.60, payment_amount=10067.60, your_number='5506715023835136')

        paymentANF = Payment(type='nf', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, pay_date='10052018', ispb_code='90400888', payment_amount=10067.60, dv='4', due_rule='7524', amount=10067.60, free_field='9485239700000007661190101', due_date='14052018', title_amount=10067.60, your_number='5511972063805440', nf_document='5646510065784320', favored_document_number='11111111111111')

        # add payments
        payment_file.add_payment(paymentA)
        payment_file.add_payment(paymentJ)
        payment_file.add_payment(paymentANF)

        # generate
        payment_file.generate()

        body_big_line = None
        body_big_line = ''.join(payment_file.body)

        root_path = os.path.dirname(os.path.abspath(__file__))
        body_big_file_path = os.path.join(root_path, 'body_big_file_multiples_segments.txt')
        f = open(body_big_file_path, 'w')
        f.write(body_big_line)
        f.close()

        assert '34100011C2041040 207179434000140                    01111 000011111111 3BF SERVICOS DE COBRANCA LTDA                                          AV ANDROMEDA                  02000BL84 ANDAR     BARUERI             06473000SP                  3410001300001A00000003301111 000011111111 0CLIENTE TESTE                 5511972063805440    10052018REA904008880000000000000000240000                            000000000000000                    00000011111111111111                       34100015         000003000000000000240000000000000000000000                                                                                                                                                                                     34100021C2031030 207179434000140                    01111 000011111111 3BF SERVICOS DE COBRANCA LTDA                                          AV ANDROMEDA                  02000BL84 ANDAR     BARUERI             06473000SP                  3410002300001J00003394752400010067609485239700000007661190101CLIENTE TESTE                 14052018000000001006760000000000000000000000000000000100520180000000010067600000000000000005506715023835136                                          34100025         000003000000000001006760000000000000000000                                                                                                                                                                                     34100031C2041040 207179434000140                    01111 000011111111 3BF SERVICOS DE COBRANCA LTDA                                          AV ANDROMEDA                  02000BL84 ANDAR     BARUERI             06473000SP                  3410003300001A00000003301111 000011111111 0CLIENTE TESTE                 5511972063805440    10052018REA000000000000000000000001006760                            00000000000000056465100657843      000000111111111111112                      34100035         000003000000000001006760000000000000000000                                                                                                                                                                                     ' in body_big_line
    
    def test_6_generate_footer(self):
        payment_file = File('itau', company)
        payment_file.header.set_company_data(company)
        payment_file.payments = []

        paymentA = Payment(type='ted', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, your_number='5511972063805440', pay_date='10052018', ispb_code='90400888', payment_amount=2400.00, favored_document_number='11111111111111')

        paymentJ = Payment(type='other_bank_slip', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, pay_date='10052018', ispb_code='90400888', currency_type=9, dv='4', due_rule='7524', amount=10067.60, free_field='9485239700000007661190101', due_date='14052018', title_amount=10067.60, payment_amount=10067.60, your_number='5506715023835136')

        paymentANF = Payment(type='nf', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, pay_date='10052018', ispb_code='90400888', payment_amount=10067.60, dv='4', due_rule='7524', amount=10067.60, free_field='9485239700000007661190101', due_date='14052018', title_amount=10067.60, your_number='5511972063805440', nf_document='5646510065784320', favored_document_number='11111111111111')

        # add payment
        payment_file.add_payment(paymentA)
        payment_file.add_payment(paymentJ)
        payment_file.add_payment(paymentANF)

        # generate
        payment_file.generate()

        footer_line = payment_file.footer.to_line()
        # print(footer_line)
        assert '34199999         000003000011                                                                                                                                                                                                                   ' in footer_line
    
    def test_7_save_file(self):
        payment_file = File('itau', company)
        payment_file.header.set_company_data(company)
        payment_file.payments = []

        paymentA = Payment(type='ted', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, your_number='5511972063805440', pay_date='10052018', ispb_code='90400888', payment_amount=2400.00, favored_document_number='11111111111111')

        paymentJ = Payment(type='other_bank_slip', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, pay_date='10052018', ispb_code='90400888', currency_type=9, dv='4', due_rule='7524', amount=10067.60, free_field='9485239700000007661190101', due_date='14052018', title_amount=10067.60, payment_amount=10067.60, your_number='5506715023835136')

        paymentANF = Payment(type='nf', favored_name='Cliente Teste', favored_bank='033', agency='01111', account='000011111111', account_digit=0, pay_date='10052018', ispb_code='90400888', payment_amount=10067.60, dv='4', due_rule='7524', amount=10067.60, free_field='9485239700000007661190101', due_date='14052018', title_amount=10067.60, your_number='5511972063805440', nf_document='5646510065784320', favored_document_number='11111111111111')

        # add payments
        payment_file.add_payment(paymentA)
        payment_file.add_payment(paymentJ)
        payment_file.add_payment(paymentANF)

        # generate
        file_full_path = payment_file.generate(os.path.dirname(os.path.abspath(__file__)))

        f = open(file_full_path, 'r')
        file_content = f.read()
        f.close()

        assert file_content in file_content
    
    def test_8_read_header_form_file(self):
        payment_file = File('itau')
        payment_file.payments = []

        f = open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'SB23058F.RET'), 'r')
        file_content = f.read()
        f.close()

        payment_file.read_file_content(file_content)

        print(payment_file.company)
        print(payment_file.payments)
        
        assert file_content in file_content


if __name__ == '__main__':
    unittest.main()


from datetime import datetime
from pg_cnab240.file_section import FileSection
from pg_cnab240.company import Company


class FileHeader(FileSection):
    def __init__(self, data=None):
        super().__init__('FileHeader', data, {
            'bank_code': {
                'type': 'int',
                'length': 3,
                'default': 341,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 0,
                'end': 3,
                'value': None,
            },
            'lot_code': {
                'type': 'int',
                'length': 4,
                'default': 0000,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 3,
                'end': 7,
                'value': None,
            },
            'register_type': {
                'type': 'int',
                'length': 1,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 7,
                'end': 8,
                'value': None,
            },
            'register_type_whites': {
                'type': 'whites',
                'length': 6,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 8,
                'end': 14,
                'value': None,
            },
            'file_layout': {
                'type': 'int',
                'length': 3,
                'default': '081',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 14,
                'end': 17,
                'value': None,
            },
            'company_document_type': {
                'type': 'int',
                'length': 1,
                'default': 2, # 1 = CPF / 2 = CNPJ
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 17,
                'end': 18,
                'value': None,
            },
            'company_document_number': {
                'type': 'int',
                'length': 14,
                'default': '',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 18,
                'end': 32,
                'value': None,
            },
            'inscription_number_whites': {
                'type': 'whites',
                'length': 20,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 32,
                'end': 52,
                'value': None,
            },
            'agency': {
                'type': 'int',
                'length': 5,
                'default': '',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 52,
                'end': 57,
                'value': None,
            },
            'agency_whites': {
                'type': 'whites',
                'length': 1,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 57,
                'end': 58,
                'value': None,
            },
            'account': {
                'type': 'int',
                'length': 12,
                'default': '',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 58,
                'end': 70,
                'value': None,
            },
            'account_whites': {
                'type': 'whites',
                'length': 1,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 70,
                'end': 71,
                'value': None,
            },
            'dac': { # bank account digit
                'type': 'int',
                'length': 1,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 71,
                'end': 72,
                'value': None,
            },
            'company_name': {
                'type': 'string',
                'length': 30,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 72,
                'end': 102,
                'value': None,
            },
            'bank_name': {
                'type': 'string',
                'length': 30,
                'default': 'BANCO ITAU SA',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 102,
                'end': 132,
                'value': None,
            },
            'bank_name_whites': {
                'type': 'whites',
                'length': 10,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 132,
                'end': 142,
                'value': None,
            },
            'file_code': {
                'type': 'int',
                'length': 1,
                'default': 1, # 1 - Shipping / 2 - Return
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 142,
                'end': 143,
                'value': None,
            },
            'generation_date': {
                'type': 'date',
                'length': 8,
                'default': datetime.utcnow().strftime(self.default_date_format),
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 143,
                'end': 151,
                'value': None,
            },
            'generation_hour': {
                'type': 'date',
                'length': 6,
                'default': datetime.utcnow().strftime(self.default_time_format),
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 151,
                'end': 157,
                'value': None,
            },
            'generation_zeros': {
                'type': 'int',
                'length': 9,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 157,
                'end': 166,
                'value': None,
            },
            'density_unit': {
                'type': 'int',
                'length': 5,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 166,
                'end': 171,
                'value': None,
            },
            'density_unit_whites': {
                'type': 'whites',
                'length': 69,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 171,
                'end': 240,
                'value': None,
            },
        })

    def set_company_data(self, company):
        if not company:
            raise Exception('Company cannot be None')
        
        super().set_data(dict(
            company_document_type = self.bank.get_company_document_id(company.document_type),
            company_document_number = company.document,
            agency = company.bank_account.agency,
            account = company.bank_account.account,
            dac = company.bank_account.digit,
            company_name = company.name,
        ))
    
    def extract_company(self):
        company = Company(self.attributes['company_name'].value, self.attributes['company_document_number'].value)
        company.set_bank_acccount(self.attributes['bank_code'].value, self.attributes['agency'].value, self.attributes['account'].value, self.attributes['dac'].value)

        return company



from unicodedata import normalize


class Attribute:
    def __init__(self, name, type, length, start, end, default_value='', pad_content=0, pad_direction='left', required=False, datetime_format='', date_format='', time_format=''):
        self.name = name
        self.type = type
        self.length = length
        self.start = start
        self.end = end
        self.default_value = default_value
        self.pad_content = str(pad_content)
        self.pad_direction = pad_direction
        self.required = required
        self.value = None

    def is_required(self):
        return self.required
    
    def get_value(self):
        if self.value:
            return self.value
        self.set_value(self.default_value)
        return self.value
    
    def set_value(self, new_value):
        if new_value is None:
            new_value = self.default_value
        
        if self.type == 'int':
            self.value = int(new_value)
        elif self.type == 'string':
            self.value = str(new_value)
        elif self.type == 'float':
            self.value = str(round(float(new_value), 2))
            self.value = self.value.split('.')
            if len(self.value[1]) < 2:
                self.value[1] += '0'
            self.value = ''.join(self.value)
        elif self.type == 'whites':
            self.value = ' '
        elif self.type == 'zeros':
            self.value = '0'
        elif self.type == 'date':
            self.value = str(new_value)
        
        self.value = str(self.value)

        self.clean_value()
        
        self.pad_value()

        self.value = self.value.upper() # transform to upper case
    
    def clean_value(self):
        # remove special chars
        special_chars = [',', '.', '-', '_', '*', '&', '´', '`', "'", '"', '!', '?', '/', ':', ';', '>', '<', '^', '~', ']', '}', '{', '[', 'ª', 'º', '#', '%', '¨', '(', ')', '+', '=', '§', '¬', '£', '¹', '²', '³', '°']
        for char in special_chars:
            self.value = self.value.replace(char, '')
        
        # remove accents
        self.value = normalize('NFKD', self.value).encode('ASCII', 'ignore').decode('ASCII')
    
    def pad_value(self):
        if len(self.value) < self.length:
            if self.pad_direction == 'left':
                self.value = self.value.rjust(self.length, self.pad_content)
            else:
                self.value = self.value.ljust(self.length, self.pad_content)
        elif len(self.value) > self.length:
            self.value = self.value[:self.length]



import json
from pg_cnab240.attribute import Attribute


class FileSection:
    default_date_format = "%d%m%Y"
    default_datetime_format = "%d%m%Y %H%M%S"
    default_time_format = "%H%M%S"

    def __init__(self, section_name, data, attributes):
        self.bank = None
        self.section_name = section_name
        self.attributes = attributes
        
        self.data = data
        if self.data is None:
            self.data = dict()

        self.transform_attributes()
        if self.data is not None:
            self.associate_data()
    
    def transform_attributes(self):
        for attr, data in self.attributes.items():
            self.attributes[attr] = Attribute(attr, data['type'], data['length'], data['start'], data['end'], data['default'], data['pad_content'], data['pad_direction'], data['required'], self.default_datetime_format, self.default_date_format, self.default_time_format)
    
    def associate_data(self):
        if self.data:
            if self.bank:
                self.data['bank_code'] = self.bank.code
             
            for name, attr in self.attributes.items():
                if name in self.data:
                    self.attributes[name].set_value(self.data[name])
                elif attr.is_required():
                    raise Exception('The ' + self.section_name + ' Attribute "' + name + '" is required')
    
    def get_dict(self):
        response = dict()
        for attr_name, attr in self.attributes.items():
            response[attr_name] = attr.get_value()
        return response
    
    def get_json(self):
        return json.dumps(self.get_dict())
    
    def to_line(self):
        line = ''
        for attr_name, attr in self.attributes.items():
            line += attr.get_value()
        return line
    
    def get_required_attributes(self):
        required_attributes = []
        for attr_name, attr in self.attributes.items():
            if attr.is_required():
                required_attributes.append(attr)
        return required_attributes
    
    def set_bank(self, bank):
        self.bank = bank
    
    def set_data(self, data=dict()):
        for attribute, value in data.items():
            self.data[attribute] = value
        self.associate_data()
    
    def set_attributes_from_line(self, line=""):
        for attr_name, attr in self.attributes.items():
            self.attributes[attr_name].set_value(line[attr.start:attr.end])
            self.attributes[attr_name].value = self.attributes[attr_name].value.strip()



from datetime import datetime
from pg_cnab240.segment_section import SegmentSection
from pg_cnab240.banks.itau.segments.headers.slip_header import SlipHeader
from pg_cnab240.banks.itau.segments.footers.slip_footer import SlipFooter


class SegmentJ(SegmentSection):
    def __init__(self, data=None):
        super().__init__('SegmentJ', data, {
            'bank_code': {
                'type': 'int',
                'length': 3,
                'default': 341,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 0,
                'end': 3,
                'value': None,
            },
            'lot_code': {
                'type': 'int',
                'length': 4,
                'default': 0000,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 3,
                'end': 7,
                'value': None,
            },
            'register_type': {
                'type': 'int',
                'length': 1,
                'default': 3,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 7,
                'end': 8,
                'value': None,
            },
            'register_number': {
                'type': 'int',
                'length': 5,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 8,
                'end': 13,
                'value': None,
            },
            'segment': {
                'type': 'string',
                'length': 1,
                'default': 'J',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 13,
                'end': 14,
                'value': None,
            },
            'move_type': {
                'type': 'int',
                'length': 3,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 14,
                'end': 17,
                'value': None,
            },
            'favored_bank': {
                'type': 'int',
                'length': 3,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 17,
                'end': 20,
                'value': None,
            },
            'currency_type': {
                'type': 'int',
                'length': 1,
                'default': 9,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 20,
                'end': 21,
                'value': None,
            },
            'dv': {
                'type': 'int',
                'length': 1,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 21,
                'end': 22,
                'value': None,
            },
            'due_rule': {
                'type': 'int',
                'length': 4,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 22,
                'end': 26,
                'value': None,
            },
            'amount': {
                'type': 'float',
                'length': 10,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 26,
                'end': 36,
                'value': None,
            },
            'free_field': {
                'type': 'int',
                'length': 25,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 36,
                'end': 61,
                'value': None,
            },
            'favored_name': {
                'type': 'string',
                'length': 30,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 62,
                'end': 91,
                'value': None,
            },
            'due_date': {
                'type': 'date',
                'length': 8,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 91,
                'end': 99,
                'value': None,
            },
            'title_amount': {
                'type': 'float',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 99,
                'end': 114,
                'value': None,
            },
            'discounts': {
                'type': 'float',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 114,
                'end': 129,
                'value': None,
            },
            'additions': { # forfeit + interest amounts
                'type': 'float',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 129,
                'end': 144,
                'value': None,
            },
            'pay_date': {
                'type': 'date',
                'length': 8,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 144,
                'end': 152,
                'value': None,
            },
            'payment_amount': {
                'type': 'float',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 152,
                'end': 167,
                'value': None,
            },
            'payment_zeros': {
                'type': 'zeros',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 167,
                'end': 182,
                'value': None,
            },
            'your_number': {
                'type': 'string',
                'length': 20,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 182,
                'end': 202,
                'value': None,
            },
            'your_number_whites': {
                'type': 'whites',
                'length': 13,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 202,
                'end': 215,
                'value': None,
            },
            'our_number': {
                'type': 'string',
                'length': 15,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 215,
                'end': 230,
                'value': None,
            },
            'occurrences': {
                'type': 'string',
                'length': 10,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 230,
                'end': 240,
                'value': None,
            },
        }, SlipHeader, SlipFooter)



from datetime import datetime
from pg_cnab240.segment_section import SegmentSection


class SlipHeader(SegmentSection):
    def __init__(self, data=None):
        super().__init__('SlipHeader', data, {
            'bank_code': {
                'type': 'int',
                'length': 3,
                'default': 341,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 0,
                'end': 3,
                'value': None,
            },
            'lot_code': {
                'type': 'int',
                'length': 4,
                'default': 0000,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 3,
                'end': 7,
                'value': None,
            },
            'register_type': {
                'type': 'int',
                'length': 1,
                'default': 1,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 7,
                'end': 8,
                'value': None,
            },
            'operation_type': {
                'type': 'string',
                'length': 1,
                'default': 'C',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 8,
                'end': 9,
                'value': None,
            },
            'payment_type': {
                'type': 'int',
                'length': 2,
                'default': 20,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 9,
                'end': 11,
                'value': None,
            },
            'payment_way': {
                'type': 'int',
                'length': 2,
                'default': 31,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 11,
                'end': 13,
                'value': None,
            },
            'lot_layout': {
                'type': 'int',
                'length': 3,
                'default': '030',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 13,
                'end': 16,
                'value': None,
            },
            'lot_layout_whites': {
                'type': 'whites',
                'length': 1,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 16,
                'end': 17,
                'value': None,
            },
            'company_document_type': {
                'type': 'int',
                'length': 1,
                'default': 2, # 1 = CPF / 2 = CNPJ
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 17,
                'end': 18,
                'value': None,
            },
            'company_document_number': {
                'type': 'int',
                'length': 14,
                'default': '',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 18,
                'end': 32,
                'value': None,
            },
            'company_document_whites': {
                'type': 'whites',
                'length': 20,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 32,
                'end': 52,
                'value': None,
            },
            'agency': {
                'type': 'int',
                'length': 5,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 52,
                'end': 57,
                'value': None,
            },
            'agency_whites': {
                'type': 'whites',
                'length': 1,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 57,
                'end': 58,
                'value': None,
            },
            'account': {
                'type': 'int',
                'length': 12,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 58,
                'end': 70,
                'value': None,
            },
            'account_whites': {
                'type': 'whites',
                'length': 1,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 70,
                'end': 71,
                'value': None,
            },
            'dac': {
                'type': 'int',
                'length': 1,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 71,
                'end': 72,
                'value': None,
            },
            'company_name': {
                'type': 'string',
                'length': 30,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 72,
                'end': 102,
                'value': None,
            },
            'lot_goal': {
                'type': 'string',
                'length': 30,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 102,
                'end': 132,
                'value': None,
            },
            'account_history': {
                'type': 'string',
                'length': 10,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 132,
                'end': 142,
                'value': None,
            },
            'company_address_street': {
                'type': 'string',
                'length': 30,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 142,
                'end': 172,
                'value': None,
            },
            'company_address_number': {
                'type': 'int',
                'length': 5,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 172,
                'end': 177,
                'value': None,
            },
            'company_address_complement': {
                'type': 'string',
                'length': 15,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 177,
                'end': 192,
                'value': None,
            },
            'company_address_city': {
                'type': 'string',
                'length': 20,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 192,
                'end': 212,
                'value': None,
            },
            'company_address_zipcode': {
                'type': 'int',
                'length': 8,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 212,
                'end': 220,
                'value': None,
            },
            'company_address_state': {
                'type': 'string',
                'length': 2,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 220,
                'end': 222,
                'value': None,
            },
            'company_address_whites': {
                'type': 'whites',
                'length': 8,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 222,
                'end': 230,
                'value': None,
            },
            'occurrences': {
                'type': 'string',
                'length': 10,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 230,
                'end': 240,
                'value': None,
            },
        })



class Bank:
    def __init__(self, name, slug, code, segment_position_identifier=None, segment_identifier_length=1, segment_header_identifier_name=None):
        self.name = name
        self.slug = slug
        self.code = code
        self.available_segments = dict()
        self.segment_position_identifier = segment_position_identifier
        self.segment_identifier_length = segment_identifier_length
        self.segment_header_identifier_name = segment_header_identifier_name
    
    def set_segment(self, segment_name, segment_class, payment_types=dict(), segment_alias=None):
        params = dict(
            segment_name = segment_name,
            segment_class = segment_class,
            payment_types = payment_types,
        )

        if segment_alias:
            params['segment_alias'] = segment_alias
        else:
            params['segment_alias'] = segment_name
        
        self.available_segments[segment_name] = params
    
    def get_file_header(self):
        pass
    
    def get_file_footer(self):
        pass

    def get_segment(self, segment):
        pass
    
    def get_payment_segment(self, payment_type):
        for segment_name, segment_dict in self.available_segments.items():
            if payment_type in [*segment_dict['payment_types']]:
                return segment_dict
        raise Exception('Payment Type not Found')
    
    def identify_payment_segment(self, payment_line, payment_header_line):
        # check segment position identifier and segment header identifier name
        if not self.segment_position_identifier or not self.segment_header_identifier_name:
            return None
        
        # for each available segment...
        for segment_name, segment_data in self.available_segments.items():
            # ... check if payment line has segment alias
            if payment_line[self.segment_position_identifier:(self.segment_position_identifier + self.segment_identifier_length)] == segment_data['segment_alias']:
                # get segment class
                segment_class = segment_data['segment_class']

                # get segment class header
                segment_header = segment_class().header

                # fill segment header from payment header line
                segment_header.set_attributes_from_line(payment_header_line)

                for type_name, type_value in segment_data['payment_types'].items():
                    if type_value is None:
                        return segment_data
                    elif type_value == segment_header.attributes[self.segment_header_identifier_name].value:
                        return segment_data



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pg_cnab240",
    version="1.0.2",
    author="Banco de Formaturas",
    author_email="alamo.saravali@bancodeformaturas.com.br",
    description="A CNAB 240 Package to process Bank payment shipping files and return files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BancoDeFormaturas/python-pg-cnab240",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)


from pg_cnab240.file_section import FileSection


class SegmentSection(FileSection):
    def __init__(self, section_name, data, attributes, header_class=None, footer_class=None):
        super().__init__(section_name, data, attributes)
        if header_class:
            self.header = header_class()
        if footer_class:
            self.footer = footer_class()
        self.company = None
    
    def set_company(self, company):
        self.company = company
        self.data['company_document_type'] = self.bank.get_company_document_id(company.document_type)
        self.data['company_document_number'] = company.document
        self.data['agency'] = company.bank_account.agency
        self.data['account'] = company.bank_account.account
        self.data['dac'] = company.bank_account.digit
        self.data['company_name'] = company.name
        self.data['company_address_street'] = company.street
        self.data['company_address_number'] = company.number
        self.data['company_address_complement'] = company.complement
        self.data['company_address_city'] = company.city
        self.data['company_address_zipcode'] = company.zipcode
        self.data['company_address_state'] = company.state
    
    def set_header(self, header):
        self.header = header
        self.header.set_bank(self.bank)
        self.header.set_company(self.company)
        self.header.set_data(self.data)
        
    def set_header_data(self, header_data=dict()):
        if not self.header:
            raise Exception('Header is invalid')
        self.header.set_data(header_data)
    
    def get_header_line(self):
        if not self.header.data:
            self.set_header(self.header)
        return self.header.to_line()
    
    def set_footer(self, footer):
        self.footer = footer
        self.footer.set_bank(self.bank)
        self.footer.set_company(self.company)
        self.footer.set_data(self.data)
    
    def set_footer_data(self, footer_data=dict()):
        if not self.footer:
            raise Exception('Footer is invalid')
        self.footer.associate_data(footer_data)
    
    def get_footer_line(self):
        if not self.footer.data:
            self.set_footer(self.footer)
        return self.footer.to_line()



from datetime import datetime
from pg_cnab240.segment_section import SegmentSection


class TransferFooter(SegmentSection):
    def __init__(self, data=None):
        super().__init__('TransferFooter', data, {
            'bank_code': {
                'type': 'int',
                'length': 3,
                'default': 341,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 0,
                'end': 3,
                'value': None,
            },
            'lot_code': {
                'type': 'int',
                'length': 4,
                'default': 0000,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 3,
                'end': 7,
                'value': None,
            },
            'register_type': {
                'type': 'int',
                'length': 1,
                'default': 5,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 7,
                'end': 8,
                'value': None,
            },
            'register_whites': {
                'type': 'whites',
                'length': 9,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 8,
                'end': 17,
                'value': None,
            },
            'registers_quantity': {
                'type': 'int',
                'length': 6,
                'default': 3,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 17,
                'end': 23,
                'value': None,
            },
            'total_amount': {
                'type': 'float',
                'length': 18,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 23,
                'end': 41,
                'value': None,
            },
            'total_amount_zeros': {
                'type': 'zeros',
                'length': 18,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'right',
                'required': False,
                'start': 41,
                'end': 59,
                'value': None,
            },
            'total_amount_whites': {
                'type': 'whites',
                'length': 171,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 59,
                'end': 230,
                'value': None,
            },
            'occurrences': {
                'type': 'string',
                'length': 10,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 230,
                'end': 240,
                'value': None,
            },
        })
    
    def set_data(self, data=dict()):
        if data is None:
            data = dict()

        # set total_amount
        data['total_amount'] = data['payment_amount']

        super().set_data(data)



from datetime import datetime
from pg_cnab240.segment_section import SegmentSection
from pg_cnab240.banks.itau.segments.headers.transfer_header import TransferHeader
from pg_cnab240.banks.itau.segments.footers.transfer_footer import TransferFooter


class SegmentA(SegmentSection):
    def __init__(self, data=None):
        super().__init__('SegmentA', data, {
            'bank_code': {
                'type': 'int',
                'length': 3,
                'default': 341,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 0,
                'end': 3,
                'value': None,
            },
            'lot_code': {
                'type': 'int',
                'length': 4,
                'default': 0000,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 3,
                'end': 7,
                'value': None,
            },
            'register_type': {
                'type': 'int',
                'length': 1,
                'default': 3,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 7,
                'end': 8,
                'value': None,
            },
            'register_number': {
                'type': 'int',
                'length': 5,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 8,
                'end': 13,
                'value': None,
            },
            'segment': {
                'type': 'string',
                'length': 1,
                'default': 'A',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 13,
                'end': 14,
                'value': None,
            },
            'move_type': {
                'type': 'int',
                'length': 3,
                'default': '000',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 14,
                'end': 17,
                'value': None,
            },
            'assembly': {
                'type': 'int',
                'length': 3,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 17,
                'end': 20,
                'value': None,
            },
            'favored_bank': {
                'type': 'int',
                'length': 3,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 20,
                'end': 23,
                'value': None,
            },
            'agency_account_digit': {
                'type': 'string',
                'length': 20,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 23,
                'end': 43,
                'value': None,
            },
            'favored_name': {
                'type': 'string',
                'length': 30,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 43,
                'end': 73,
                'value': None,
            },
            'your_number': {
                'type': 'string',
                'length': 20,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 73,
                'end': 93,
                'value': None,
            },
            'pay_date': {
                'type': 'date',
                'length': 8,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 93,
                'end': 101,
                'value': None,
            },
            'currency_type': {
                'type': 'string',
                'length': 3,
                'default': 'REA',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 101,
                'end': 104,
                'value': None,
            },
            'ispb_code': {
                'type': 'int',
                'length': 8,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 104,
                'end': 112,
                'value': None,
            },
            'ispb_zeros': {
                'type': 'int',
                'length': 7,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 112,
                'end': 119,
                'value': None,
            },
            'payment_amount': {
                'type': 'float',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 119,
                'end': 134,
                'value': None,
            },
            'our_number': {
                'type': 'string',
                'length': 15,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 134,
                'end': 149,
                'value': None,
            },
            'our_number_whites': {
                'type': 'whites',
                'length': 5,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 149,
                'end': 154,
                'value': None,
            },
            'effective_pay_date': {
                'type': 'date',
                'length': 8,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 154,
                'end': 162,
                'value': None,
            },
            'effective_paid_amount': {
                'type': 'float',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 162,
                'end': 177,
                'value': None,
            },
            'goal_detail': {
                'type': 'string',
                'length': 18,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 177,
                'end': 195,
                'value': None,
            },
            'goal_whites': {
                'type': 'whites',
                'length': 2,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 195,
                'end': 197,
                'value': None,
            },
            'payment_document_number': {
                'type': 'int',
                'length': 6,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 197,
                'end': 203,
                'value': None,
            },
            'favored_document_number': {
                'type': 'int',
                'length': 14,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 203,
                'end': 217,
                'value': None,
            },
            'doc_goal': {
                'type': 'string',
                'length': 2,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 217,
                'end': 219,
                'value': None,
            },
            'ted_goal': {
                'type': 'string',
                'length': 5,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 219,
                'end': 224,
                'value': None,
            },
            'ted_whites': {
                'type': 'whites',
                'length': 5,
                'default': ' ',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 224,
                'end': 229,
                'value': None,
            },
            'favored_message': {
                'type': 'string',
                'length': 1,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 229,
                'end': 230,
                'value': None,
            },
            'occurrences': {
                'type': 'string',
                'length': 10,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 230,
                'end': 240,
                'value': None,
            },
        }, TransferHeader, TransferFooter)
    
    def set_data(self, data=dict()):
        if data is not None:
            if 'agency' in data and 'account' in data and 'account_digit' in data:
                data['agency_account_digit'] = str(data['agency']) + ' ' + str(data['account']) + ' ' + str(data['account_digit'])
        
        super().set_data(data)


class SegmentANF(SegmentSection):
    def __init__(self, data=None):
        super().__init__('SegmentANF', data, {
            'bank_code': {
                'type': 'int',
                'length': 3,
                'default': 341,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 0,
                'end': 3,
                'value': None,
            },
            'lot_code': {
                'type': 'int',
                'length': 4,
                'default': 0000,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 3,
                'end': 7,
                'value': None,
            },
            'register_type': {
                'type': 'int',
                'length': 1,
                'default': 3,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 7,
                'end': 8,
                'value': None,
            },
            'register_number': {
                'type': 'int',
                'length': 5,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 8,
                'end': 13,
                'value': None,
            },
            'segment': {
                'type': 'string',
                'length': 1,
                'default': 'A',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 13,
                'end': 14,
                'value': None,
            },
            'move_type': {
                'type': 'int',
                'length': 3,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 14,
                'end': 17,
                'value': None,
            },
            'move_type_zeros': {
                'type': 'zeros',
                'length': 3,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 17,
                'end': 20,
                'value': None,
            },
            'favored_bank': {
                'type': 'int',
                'length': 3,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 20,
                'end': 23,
                'value': None,
            },
            'agency_account_digit': {
                'type': 'string',
                'length': 20,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 23,
                'end': 43,
                'value': None,
            },
            'favored_name': {
                'type': 'string',
                'length': 30,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 43,
                'end': 73,
                'value': None,
            },
            'your_number': {
                'type': 'string',
                'length': 20,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 73,
                'end': 93,
                'value': None,
            },
            'pay_date': {
                'type': 'date',
                'length': 8,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': True,
                'start': 93,
                'end': 101,
                'value': None,
            },
            'currency_type': {
                'type': 'string',
                'length': 3,
                'default': 'REA',
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 101,
                'end': 104,
                'value': None,
            },
            'currency_zeros': {
                'type': 'zeros',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 104,
                'end': 119,
                'value': None,
            },
            'payment_amount': {
                'type': 'float',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 119,
                'end': 134,
                'value': None,
            },
            'our_number': {
                'type': 'string',
                'length': 15,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 134,
                'end': 149,
                'value': None,
            },
            'our_number_whites': {
                'type': 'whites',
                'length': 5,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 149,
                'end': 154,
                'value': None,
            },
            'effective_pay_date': {
                'type': 'date',
                'length': 8,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 154,
                'end': 162,
                'value': None,
            },
            'effective_paid_amount': {
                'type': 'float',
                'length': 15,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 162,
                'end': 177,
                'value': None,
            },
            'nf_document': {
                'type': 'int',
                'length': 14,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 177,
                'end': 191,
                'value': None,
            },
            'nf_document_whites': {
                'type': 'whites',
                'length': 6,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 191,
                'end': 197,
                'value': None,
            },
            'payment_document_number': { # doc/ted/io/cheque return number
                'type': 'int',
                'length': 6,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 197,
                'end': 203,
                'value': None,
            },
            'favored_document_number': {
                'type': 'int',
                'length': 14,
                'default': 0,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 203,
                'end': 217,
                'value': None,
            },
            'identify_type': {
                'type': 'int',
                'length': 1,
                'default': 2,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 217,
                'end': 218,
                'value': None,
            },
            'identify_type_whites': {
                'type': 'whites',
                'length': 11,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 218,
                'end': 229,
                'value': None,
            },
            'favored_message': {
                'type': 'string',
                'length': 1,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 229,
                'end': 230,
                'value': None,
            },
            'occurrences': {
                'type': 'string',
                'length': 10,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 230,
                'end': 240,
                'value': None,
            },
        }, TransferHeader, TransferFooter)
    
    def set_data(self, data=dict()):
        if data is not None:
            if 'agency' in data and 'account' in data and 'account_digit' in data:
                data['agency_account_digit'] = str(data['agency']) + ' ' + str(data['account']) + ' ' + str(data['account_digit'])
        
        super().set_data(data)



import json


class Payment:
    def __init__(self, **kwargs):
        self.attributes = dict(
            type = None,
            pay_date = None,
            effective_pay_date = None,
            favored_name = None,
            favored_document_number = None,
            favored_bank = None,
            agency = None,
            account = None,
            account_digit = None,
            our_number = None,
            currency_type = 'REA',
            your_number = None,
            ispb_code = None,
            goal_detail = None,
            payment_document_number = None,
            doc_goal = None,
            ted_goal = None,
            nf_document = None,
            identify_type = 2,
            barcode = None,
            dv = None,
            due_rule = None,
            amount = None,
            free_field = None,
            due_date = None,
            title_amount = None,
            discounts = None,
            additions = None,
            payment_amount = None,
            favored_message = None,
            occurrences = None,
            move_type = None,
            assembly = None,
        )

        kwargs_keys = kwargs.keys()
        for attr_name in self.attributes.keys():
            if attr_name in kwargs_keys:
                self.attributes[attr_name] = kwargs[attr_name]
    
    def get_attribute(self, attr_name):
        if attr_name in self.attributes.keys():
            return self.attributes[attr_name]
        raise Exception('Payment does not have attribute called "' + attr_name + '"')
    
    def set_attribute(self, attr_name, attr_value):
        self.attributes[attr_name] = attr_value
    
    def get_attributes(self):
        return self.attributes
    
    def get_dict(self):
        return self.get_attributes()
    
    def get_json(self):
        return json.dumps(self.get_attributes())



from pg_cnab240.banks.bank_account import BankAccount


class Company:
    def __init__(self, name, document):
        self.name = name
        
        self.document_type = self.get_company_document_type(document)
        self.document = document

        self.bank_account = None

        # address
        self.street = None
        self.number = None
        self.complement = None
        self.district = None
        self.city = None
        self.state = None
        self.zipcode = None
        self.country = None
    
    def set_bank_acccount(self, bank_code, agency, account, account_digit):
        self.bank_account = BankAccount(bank_code, agency, account, account_digit)
    
    def set_address(self, street, number, complement, district, city, state, zipcode=None, country='BR'):
        self.street = street
        self.number = number
        self.complement = complement
        self.district = district
        self.city = city
        self.state = state
        self.zipcode = zipcode
        self.country = country
    
    def get_company_document_type(self, document_number):
        if len(str(document_number)) == 14:
            return 'cnpj'
        return 'cpf'



from datetime import datetime
from pg_cnab240.file_section import FileSection


class FileFooter(FileSection):
    def __init__(self, data=None):
        super().__init__('FileFooter', data, {
            'bank_code': {
                'type': 'int',
                'length': 3,
                'default': 341,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 0,
                'end': 3,
                'value': None,
            },
            'lot_code': {
                'type': 'int',
                'length': 4,
                'default': 9999,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 3,
                'end': 7,
                'value': None,
            },
            'register_type': {
                'type': 'int',
                'length': 1,
                'default': 9,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 7,
                'end': 8,
                'value': None,
            },
            'register_type_whites': {
                'type': 'whites',
                'length': 9,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'left',
                'required': False,
                'start': 8,
                'end': 17,
                'value': None,
            },
            'lots_quantity': {
                'type': 'int',
                'length': 6,
                'default': 1,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': True,
                'start': 17,
                'end': 23,
                'value': None,
            },
            'registers_quantity': {
                'type': 'int',
                'length': 6,
                'default': 3,
                'pad_content': 0,
                'pad_direction': 'left',
                'required': False,
                'start': 23,
                'end': 29,
                'value': None,
            },
            'registers_quantity_whites': {
                'type': 'whites',
                'length': 211,
                'default': '',
                'pad_content': ' ',
                'pad_direction': 'right',
                'required': False,
                'start': 29,
                'end': 240,
                'value': None,
            },
        })
